# $Id: proxy.py,v 1.14 2011/05/23 20:54:12 edwlan Exp $
import copy
import urllib
import urlparse
import itertools
import traceback
import random
import time
import UserDict, collections
collections.Mapping.register(UserDict.DictMixin)

from hashlib import sha1
import emen2.util.jsonutil


class NewStyleBaseException(Exception):
    def _get_message(self):
        return self._message
    def _set_message(self, message):
        self._message = message

    message = property(_get_message, _set_message)


class JSONRPCException(NewStyleBaseException):
	def __init__(self, rpcError):
		Exception.__init__(self, rpcError.get('message'))
		self.data = rpcError.get('data')
		self.message = rpcError.get('message')
		self.code = rpcError.get('code')


inst = lambda x:x()
class JSONRPCProxy(object):

	@inst
	class IDGen(object):
		def __init__(self):
			self._hasher = sha1()
			self._id = 0
		def __get__(self, *_, **__):
			self._id += 1
			self._hasher.update(str(self._id))
			self._hasher.update(time.ctime())
			self._hasher.update(str(random.random))
			return self._hasher.hexdigest()


	@classmethod
	def from_url(cls, url, ctxid=None, serviceName=None):
		urlsp = urlparse.urlsplit(url)
		url = '%s://%s' % (urlsp.scheme, urlsp.netloc)
		path = urlsp.path
		if urlsp.query: path = '%s?%s' % (path, urlsp.query)
		if urlsp.fragment: path = '%s#%s' % (path, urlsp.fragment)
		return cls(url, path, serviceName, ctxid)


	def transform_url(*a):
		return a


	def __init__(self, host, path='/jsonrpc', serviceName=None, *args, **kwargs):
		self.serviceURL = host
		self._serviceName = serviceName
		self._path = path
		self.init(*args, **kwargs)
		self.serviceURL, self._path = self._transformURL(host, path)


	def _transformURL(self, serviceURL, path):
		if serviceURL[-1] == '/':
			serviceURL = serviceURL[:-1]
		if path[0] != '/':
			path = '/%s'%path
		if path[-1] != '/' and '?' not in path:
			path = '%s/'%path
		return serviceURL, path


	def init(self, *args, **kwargs):
		pass


	def __getattr__(self, name):
		if self._serviceName != None:
			name = "%s.%s" % (self._serviceName, name)
		return self.__class__(self.serviceURL, path=self._path, serviceName=name)


	def _get_postdata(self, args, kwargs):
		if kwargs.has_key('__args'):
			raise ValueError, 'invalid argument name: __args'
		kwargs['__args'] = args or ()
		postdata = emen2.util.jsonutil.encode({
			"method": self._serviceName,
			'params': kwargs,
			'id': self.IDGen,
			'jsonrpc': '2.0'
		})
		return postdata


	def _proc_response(self, data):
		return data


	def __call__(self, *args, **kwargs):
		print "-> ", self._serviceName
		
		url = '%(host)s%(path)s' % dict(host = self.serviceURL, path = self._path)
		postdata = self._get_postdata(args, kwargs)
		respdata = urllib.urlopen(url, postdata).read()
		resp = emen2.util.jsonutil.decode(respdata)

		if resp.get('error') != None:
			raise JSONRPCException(resp['error'])
		else:
			resp = self._proc_response(resp)
			result = resp['result']
			return result


	def _call(self, method, *args, **kwargs):
		p = self.__class__(self.serviceURL, path=self._path, serviceName=method)
		return p(*args, **kwargs)


	def batch_call(self, names, *params):
		methods = ( (getattr(self, name),param) for name,param in itertools.izip(names, params) )
		data = (method._get_postdata(params) for method, params in methods)
		postdata = '[%s]' % ','.join(data)
		respdata = urllib.urlopen(self.serviceURL, postdata).read()
		resp = emen2.util.jsonutil.decode(respdata)
		return [res.get('result', res.get('error')) for res in resp]


__version__ = "$Revision: 1.14 $".split(":")[1][:-1].strip()
