# $Id: server.py,v 1.6 2011/05/23 20:54:12 edwlan Exp $
import cgi
import time
import emen2.util.jsonutil

# Twisted imports
from twisted.web import server
from twisted.internet import threads
from twisted.web.resource import Resource


import UserDict, collections
collections.Mapping.register(UserDict.DictMixin)


## Base class providing a JSON-RPC 2.0 implementation with 2 customizable hooks
class jsonrpc(Resource):
	isLeaf = True

	def parse_data(self, content):
		if content.get('jsonrpc') != '2.0': raise ValueError, 'wrong JSON-RPC version'
		method = content.get('method')
		kwargs = content.get('params', {})
		args = ()
		if not isinstance(kwargs, dict):
			args = tuple(kwargs)
			kwargs = {}
		else:
			args = kwargs.pop('__args', args)
		kwargs = dict( (str(k), v) for k,v in kwargs.items() )
		return method, kwargs, args


	def _cbRender(self, result, request):
		request.setHeader("content-type", 'application/json')
		request.setResponseCode(200)
		result = emen2.util.jsonutil.encode(result).encode('utf-8')
		request.setHeader("content-length", len(result))
		request.write(result)
		request.finish()


	def _logerror(self, result, request):
		pass


	def _ebRender(self, result, request, content, *args, **kwargs):
		result_template = dict(
			jsonrpc='2.0',
			error= dict(
				code=0,
				message='Stub Message',
				data = 'Stub Data'
		))

		try:
			request.setHeader("X-Error", result.getErrorMessage())
			result_template['error'].update(
					code=0,
					message=' '.join(str(x) for x in result.value),
					data=content
			)
			if isinstance(content, list):
				result_template['id'] = content[0].get('id', '<NULL>')
				result_template = [result]
			else: result_template['id'] = content.get('id')
		except Exception, e:
			result_template['error'].update(
				code = 0,
				message = 'Error in errorpage: %s' % e,
				data = ''
			)

		result = result_template
		result['error']['message'] = cgi.escape(result['error']['message'])
		result = emen2.util.jsonutil.encode(result)
		result = result.encode('utf-8')
		request.setHeader("content-length", len(result))
		request.setResponseCode(500)
		self._logerror(result, request)
		request.write(result)
		request.finish()


	def render(self, request):
		ctxid = request.getCookie("ctxid") or request.args.get("ctxid", [None])[0]
		host = request.getClientIP()

		request.content.seek(0, 0)
		content = emen2.util.jsonutil.decode(request.content.read())
		d = threads.deferToThread(self.action, request, content, ctxid=ctxid, host=host)
		d.addCallback(self._cbRender, request)
		d.addErrback(self._ebRender, request, content)
		return server.NOT_DONE_YET


	def action(self, request, contents, **kw):
		result = []
		ol = (True if isinstance(contents, list) else False)
		if not ol:
			contents = [contents]

		for content in contents:
			res = dict(
				jsonrpc = '2.0',
				id = content.get('id'),
				result = self._callmethod(request, *self.parse_data(content), **kw)
			)
			self._processresult(res, request.args)

			result.append(res)


		return ( result if ol else result[0] )


	def _callmethod(self, request, method, kwargs, args, **kw):
		return 'Test Result'


	def _processresult(self, result, args):
		pass



__version__ = "$Revision: 1.6 $".split(":")[1][:-1].strip()
